from uuid import uuid4

from orchestrator.db.models import ProductTable, SubscriptionTable

# noinspection PyProtectedMember
from orchestrator.forms import FormPage
from orchestrator.forms.validators import Choice, UniqueConstrainedList
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import done, init, step, workflow
from orchestrator.workflows.steps import resync, set_status, store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form

from gso.products.product_blocks import PhyPortCapacity
from gso.products.product_blocks.iptrunk import IptrunkType
from gso.products.product_types.device import Device
from gso.products.product_types.iptrunk import IptrunkInactive, IptrunkProvisioning
from gso.services import _ipam, provisioning_proxy
from gso.services.provisioning_proxy import await_pp_results, confirm_pp_results


def initial_input_form_generator(product_name: str) -> FormGenerator:
    # TODO: we need additional validation:
    # * interface names must be validated

    devices = {}
    for device_id, device_description in (
        SubscriptionTable.query.join(ProductTable)
        .filter(
            ProductTable.product_type == "Device",
            SubscriptionTable.status == "active",
        )
        .with_entities(SubscriptionTable.subscription_id, SubscriptionTable.description)
        .all()
    ):
        devices[str(device_id)] = device_description

    class CreateIptrunkForm(FormPage):
        class Config:
            title = product_name

        geant_s_sid: str
        iptrunk_description: str
        iptrunk_type: IptrunkType
        iptrunk_speed: PhyPortCapacity
        iptrunk_minimum_links: int

    initial_user_input = yield CreateIptrunkForm

    class AeMembersListA(UniqueConstrainedList[str]):
        min_items = initial_user_input.iptrunk_minimum_links

    DeviceEnumA = Choice("Select a device", zip(devices.keys(), devices.items()))

    class CreateIptrunkSideAForm(FormPage):
        class Config:
            title = "Provide subscription details for side A of the trunk."

        iptrunk_sideA_node_id: DeviceEnumA
        iptrunk_sideA_ae_iface: str
        iptrunk_sideA_ae_geant_a_sid: str
        iptrunk_sideA_ae_members: AeMembersListA
        iptrunk_sideA_ae_members_descriptions: AeMembersListA

    user_input_side_a = yield CreateIptrunkSideAForm

    # We remove the selected device for side A, to prevent any loops
    devices.pop(str(user_input_side_a.iptrunk_sideA_node_id.name))
    DeviceEnumB = Choice("Select a device", zip(devices.keys(), devices.items()))

    class AeMembersListB(UniqueConstrainedList[str]):
        min_items = len(user_input_side_a.iptrunk_sideA_ae_members)
        max_items = len(user_input_side_a.iptrunk_sideA_ae_members)

    class CreateIptrunkSideBForm(FormPage):
        class Config:
            title = "Provide subscription details for side B of the trunk."

        iptrunk_sideB_node_id: DeviceEnumB
        iptrunk_sideB_ae_iface: str
        iptrunk_sideB_ae_geant_a_sid: str
        iptrunk_sideB_ae_members: AeMembersListB
        iptrunk_sideB_ae_members_descriptions: AeMembersListB

    user_input_side_b = yield CreateIptrunkSideBForm

    return initial_user_input.dict() | user_input_side_a.dict() | user_input_side_b.dict()


@step("Create subscription")
def create_subscription(product: UUIDstr) -> State:
    subscription = IptrunkInactive.from_product_id(product, uuid4())

    return {
        "subscription": subscription,
        "subscription_id": subscription.subscription_id,
    }


@step("Get information from IPAM")
def get_info_from_ipam(subscription: IptrunkProvisioning) -> State:
    # TODO: get info about how these should be generated
    subscription.iptrunk.iptrunk_ipv4_network = _ipam.allocate_service_ipv4_network(
        service_type="TRUNK",
        comment=subscription.iptrunk.iptrunk_description,
    ).v4
    subscription.iptrunk.iptrunk_ipv6_network = _ipam.allocate_service_ipv6_network(
        service_type="TRUNK",
        comment=subscription.iptrunk.iptrunk_description,
    ).v6
    return {"subscription": subscription}


@step("Initialize subscription")
def initialize_subscription(
    subscription: IptrunkInactive,
    geant_s_sid: str,
    iptrunk_type: IptrunkType,
    iptrunk_description: str,
    iptrunk_speed: str,
    iptrunk_minimum_links: int,
    iptrunk_sideA_node_id: str,
    iptrunk_sideA_ae_iface: str,
    iptrunk_sideA_ae_geant_a_sid: str,
    iptrunk_sideA_ae_members: list[str],
    iptrunk_sideA_ae_members_descriptions: list[str],
    iptrunk_sideB_node_id: str,
    iptrunk_sideB_ae_iface: str,
    iptrunk_sideB_ae_geant_a_sid: str,
    iptrunk_sideB_ae_members: list[str],
    iptrunk_sideB_ae_members_descriptions: list[str],
) -> State:
    subscription.iptrunk.geant_s_sid = geant_s_sid
    subscription.iptrunk.iptrunk_description = iptrunk_description
    subscription.iptrunk.iptrunk_type = iptrunk_type
    subscription.iptrunk.iptrunk_speed = iptrunk_speed
    subscription.iptrunk.iptrunk_isis_metric = 9000
    subscription.iptrunk.iptrunk_minimum_links = iptrunk_minimum_links

    subscription.iptrunk.iptrunk_sideA_node = Device.from_subscription(iptrunk_sideA_node_id).device
    subscription.iptrunk.iptrunk_sideA_ae_iface = iptrunk_sideA_ae_iface
    subscription.iptrunk.iptrunk_sideA_ae_geant_a_sid = iptrunk_sideA_ae_geant_a_sid
    subscription.iptrunk.iptrunk_sideA_ae_members = iptrunk_sideA_ae_members
    subscription.iptrunk.iptrunk_sideA_ae_members_description = iptrunk_sideA_ae_members_descriptions

    subscription.iptrunk.iptrunk_sideB_node = Device.from_subscription(iptrunk_sideB_node_id).device
    subscription.iptrunk.iptrunk_sideB_ae_iface = iptrunk_sideB_ae_iface
    subscription.iptrunk.iptrunk_sideB_ae_geant_a_sid = iptrunk_sideB_ae_geant_a_sid
    subscription.iptrunk.iptrunk_sideB_ae_members = iptrunk_sideB_ae_members
    subscription.iptrunk.iptrunk_sideB_ae_members_description = iptrunk_sideB_ae_members_descriptions

    subscription.description = f"IP trunk, geant_s_sid:{geant_s_sid}"
    subscription = IptrunkProvisioning.from_other_lifecycle(subscription, SubscriptionLifecycle.PROVISIONING)

    return {"subscription": subscription}


@step("Provision IP trunk interface [DRY RUN]")
def provision_ip_trunk_iface_dry(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "trunk_interface")

    return {
        "subscription": subscription,
        "label_text": ("Provision of the Trunk interface [DRY] " "Please refresh to get the results of the playbook"),
    }


@step("Provision IP trunk interface [FOR REAL]")
def provision_ip_trunk_iface_real(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "trunk_interface", False)

    return {
        "subscription": subscription,
        "label_text": ("Provision of the Trunk interface [REAL] " "Please refresh to get the results of the playbook"),
    }


@step("Provision IP trunk ISIS interface [DRY RUN]")
def provision_ip_trunk_isis_iface_dry(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "isis_interface")

    return {
        "subscription": subscription,
        "label_text": ("Provision of the ISIS interface [DRY]" "Please refresh to get the results of the playbook"),
    }


@step("Provision IP trunk ISIS interface [FOR REAL]")
def provision_ip_trunk_isis_iface_real(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "isis_interface", False)

    return {
        "subscription": subscription,
        "label_text": ("Provision of the ISIS interface [REAL]" "Please refresh to get the results of the playbook"),
    }


@step("Provision IP trunk LDP interface [DRY RUN]")
def provision_ip_trunk_ldp_iface_dry(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "ldp_interface")

    return {
        "subscription": subscription,
        "label_text": ("Provision of the LDP interface [DRY]" "Please refresh to get the results of the playbook"),
    }


@step("Provision IP trunk LDP interface [FOR REAL]")
def provision_ip_trunk_ldp_iface_real(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "ldp_interface", False)

    return {
        "subscription": subscription,
        "label_text": ("Provision of the LDP interface [REAL]" "Please refresh to get the results of the playbook"),
    }


@step("Provision IP trunk LLDP interface [DRY RUN]")
def provision_ip_trunk_lldp_iface_dry(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "lldp_interface")

    return {
        "subscription": subscription,
        "label_text": ("Provision of the LLDP interface [DRY]" "Please refresh to get the results of the playbook"),
    }


@step("Provision IP trunk LLDP interface [FOR REAL]")
def provision_ip_trunk_lldp_iface_real(subscription: IptrunkProvisioning, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "lldp_interface", False)

    return {
        "subscription": subscription,
        "label_text": ("Provision of the LLDP interface [REAL]" "Please refresh to get the results of the playbook"),
    }


@workflow(
    "Create IP trunk",
    initial_input_form=wrap_create_initial_input_form(initial_input_form_generator),
    target=Target.CREATE,
)
def create_iptrunk():
    return (
        init
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> get_info_from_ipam
        >> provision_ip_trunk_iface_dry
        >> await_pp_results
        >> confirm_pp_results
        >> provision_ip_trunk_iface_real
        >> await_pp_results
        >> confirm_pp_results
        >> provision_ip_trunk_isis_iface_dry
        >> await_pp_results
        >> confirm_pp_results
        >> provision_ip_trunk_isis_iface_real
        >> await_pp_results
        >> confirm_pp_results
        >> provision_ip_trunk_ldp_iface_dry
        >> await_pp_results
        >> confirm_pp_results
        >> provision_ip_trunk_ldp_iface_real
        >> await_pp_results
        >> confirm_pp_results
        >> provision_ip_trunk_lldp_iface_dry
        >> await_pp_results
        >> confirm_pp_results
        >> provision_ip_trunk_lldp_iface_real
        >> await_pp_results
        >> confirm_pp_results
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
