from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, UUIDstr
from orchestrator.workflow import done, init, step, workflow
from orchestrator.workflows.steps import resync, store_process_subscription, unsync
from orchestrator.workflows.utils import wrap_modify_initial_input_form

from gso.products.product_types.iptrunk import Iptrunk
from gso.services import provisioning_proxy
from gso.services.provisioning_proxy import await_pp_results, confirm_pp_results


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    subscription = Iptrunk.from_subscription(subscription_id)

    class ModifyIptrunkForm(FormPage):
        isis_metric: int = subscription.iptrunk.iptrunk_isis_metric

    user_input = yield ModifyIptrunkForm

    return user_input.dict()


@step("Update subscription")
def modify_iptrunk_subscription(subscription: Iptrunk, isis_metric: int) -> State:
    subscription.iptrunk.iptrunk_isis_metric = isis_metric

    return {"subscription": subscription}


@step("Provision IP trunk ISIS interface [DRY RUN]")
def provision_ip_trunk_isis_iface_dry(subscription: Iptrunk, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "isis_interface")

    return {
        "subscription": subscription,
        "label_text": (
            "This is a dry run for the deployment of a new IP "
            "trunk ISIS interface. "
            "Deployment is being taken care of by the "
            "provisioning proxy, please wait for the results to "
            "come back before continuing."
        ),
    }


@step("Provision IP trunk ISIS interface [FOR REAL]")
def provision_ip_trunk_isis_iface_real(subscription: Iptrunk, process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id, "isis_interface", False)

    return {
        "subscription": subscription,
        "label_text": (
            "This is a live deployment of a new IP trunk "
            "ISIS interface. "
            "Deployment is being taken care of by the "
            "provisioning proxy, please wait for the results to "
            "come back before continuing."
        ),
    }


@workflow(
    "Modify IP trunk",
    initial_input_form=wrap_modify_initial_input_form(initial_input_form_generator),
    target=Target.MODIFY,
)
def modify_isis_metric():
    return (
        init
        >> store_process_subscription(Target.MODIFY)
        >> unsync
        >> modify_iptrunk_subscription
        >> provision_ip_trunk_isis_iface_dry
        >> await_pp_results
        >> confirm_pp_results
        >> provision_ip_trunk_isis_iface_real
        >> await_pp_results
        >> confirm_pp_results
        >> resync
        >> done
    )
